#include "Service.hpp"

// C++ system level
// #include <cstring>     // memset, strncmp
// #include <cstdlib>     // TODO: is this for memcpy?
// #include <functional>

// ESP32 specific
#include "esp_log.h"

// project specific
#include <Types.hpp>
#include "SiliconTorch/NVSExplorer.hpp"

// qthing stuff
// #include <qthing>

// misc
#include <nlohmann/json.hpp>


using json = nlohmann::json;


static const char* TAG = "ServiceManager";


namespace SiliconTorch {

  namespace ServiceManager {

    const str NVSStateNameSpace = "__services__";


    Service::Service() {
      // u8 state = 0;  // disabled -> 0   enabled -> 1
      // state = nvs.getUnsignedInt(nameSpace(), NVSStateNameSpace, state);

      // if (state >= (u8)Invalid) state = (u8)Disabled;
      // nvsState = (NVSState)state;
    }

    void Service::preInit(ServiceLookup lookup) {
      serviceLookup = lookup;
    }

    void Service::postInit() {

      attrsLocked = true;

      if (getName().length() < 1)
        hasErrors = true;  // service is errorneous because of unset name property


      for (const auto& name : *wantedNames) {
        Service* service = otherByName(name);

        if (service != NULL) wantedServices.insert(service);
      }

      for (const auto& name : *requiredNames) {
        Service* service = otherByName(name);

        if (service != NULL) {
          requiredServices.insert(service);
        } else {
          hasErrors = true;  // service is errorneous because of unsatisfied dependency
          ESP_LOGW(TAG, "Service[ %s ] is errorneous because of unsatisfied dependency[ %s ]", getName().c_str(), name.c_str());
        }
      }

      delete wantedNames;
      delete requiredNames;

      wantedNames = NULL;
      requiredNames = NULL;


      u8 state = 0;  // disabled -> 0   enabled -> 1
      state = SiliconTorch::NVSExplorer::NVSExplorer::instance().getUnsignedInt(NVSStateNameSpace, getName(), state);

      if (state >= (u8)Invalid) state = (u8)Disabled;
      nvsState = (NVSState)state;

    }

    bool Service::isEnabled() const {
      return nvsState == Enabled && !hasErrors;
    }

    void Service::setEnabled(bool enabled) {

      NVSState state = enabled ? Enabled : Disabled;
      u8 stateValue = (u8)state;

      bool result = SiliconTorch::NVSExplorer::NVSExplorer::instance().setU8(NVSStateNameSpace, getName(), stateValue);

      if (!result) {
        SiliconTorch::NVSExplorer::NVSExplorer::instance().removeKey(NVSStateNameSpace, getName());
        result = SiliconTorch::NVSExplorer::NVSExplorer::instance().setU8(NVSStateNameSpace, getName(), stateValue);
      }

      if (result) {
        nvsState = state;
        ESP_LOGI(TAG, "Setting service[ %s ] to %s", getName().c_str(), enabled ? "enabled[✅]" : "disabled[❌]");
      } else {
        ESP_LOGW(TAG, "Error saving service[ %s] state!", getName().c_str());
      }
    }


    void Service::setIcon(const str& icon) {
      if (!attrsLocked) {
        this->icon = icon;
      } else {
        // TODO: report error somehow?
      }
    }

    void Service::setName(const str& name) {
      if (!attrsLocked) {
        // TODO: length check!!  (reason: NVS)
        this->name = name;
      } else {
        // TODO: report error somehow?
      }
    }

    void Service::setNameSpace(const str& nameSpace) {
      if (!attrsLocked) {
        // TODO: length check!!  (reason: NVS)
        this->nameSpace = nameSpace;
      } else {
        // TODO: report error somehow?
      }
    }

    void Service::setDisplayName(const str& displayName) {
      if (!attrsLocked) {
        this->displayName = displayName;
      } else {
        // TODO: report error somehow?
      }
    }

    void Service::setIconLength(size_t iconLength) {
      if (!attrsLocked) {
        this->iconLength = iconLength;
      } else {
        // TODO: report error somehow?
      }
    }


    void Service::addWantedService(const str& name) {
      if (!attrsLocked) {
        wantedNames->insert(name);
      } else {
        // TODO: report error somehow?
      }
    }

    void Service::addRequiredService(const str& name) {
      if (!attrsLocked) {
        requiredNames->insert(name);
      } else {
        // TODO: report error somehow?
      }
    }


    ServiceSet Service::getWantedServices() {
      ServiceSet out;

      // TODO: better method for set copy?!

      for (const auto& service : wantedServices)
        out.insert(service);

      return out;
    }

    ServiceSet Service::getRequiredServices() {
      ServiceSet out;

      // TODO: better method for set copy?!

      for (const auto& service : requiredServices)
        out.insert(service);

      return out;
    }


    Service* Service::otherByName(const str& name) {
      return serviceLookup(name);
    }


    const str& Service::getIcon() const {
      return icon;
    }

    const str& Service::getName() const {
      return name;
    }

    const str& Service::getNameSpace() const {
      return nameSpace;
    }

    const str& Service::getDisplayName() const {
      if (displayName.length() > 0) return displayName;
      else                          return name;
    }

    str Service::getFancyName() const {

      str out = getDisplayName();

      if (icon.length() > 0) {
        out += str("[");
        out += icon;
        out += str("]");
      }

      return out;
    }


    size_t Service::getIconLength() const {
      return iconLength;
    }

    size_t Service::getFancyNameLength() const {
      size_t out = getDisplayName().length();

      if (iconLength > 0)
        out += iconLength;
      else if (icon.length() > 0)
        // The icon *should* contain one unicode symbol which kitty renders as 2 chars wide
        // Using icon.length() would break the rendering
        out += 2;

      // Brackets around icon if non-empty
      if (icon.length() > 0)
        out += 2;

      return out;
    }


    json Service::getConfigJSON() const {
      return "{}"_json;
    }

    json Service::getInfoJSON() const {

      json info;

      json cfg = getConfigJSON();
      if (!cfg.is_null())
        info["config"] = cfg;
      else
        info["config"] = "{}"_json;

      if (getIcon().length() > 0)
        info["Icon"] = getIcon();

      if (getNameSpace().length() > 0)
        info["NameSpace"] = getNameSpace();

      info["FancyName"] = getFancyName();
      info["DisplayName"] = getDisplayName();

      info["enabled"] = nvsState == Enabled;
      info["hasErrors"] = hasErrors;

      return info;
    }


  }
}
